#include <Arduino.h>
#include <FS.h>                   //this needs to be first, or it all crashes and burns...
#include <ESP8266WiFi.h>          //https://github.com/esp8266/Arduino

//needed for library
#include <DNSServer.h>
#include <ESP8266WebServer.h>
#include <WiFiManager.h>

#include <ArduinoJson.h>
#include <SPI.h>

#include "config.h"
#include "display.h"
#include "fastclock.h"
#include "radio.h"

#define AUTOSTART_CLOCK_AFTER_ms 10000
#define POWER_OFF_PIN 0

//define your default values here, if there are different values in config.json, they are overwritten.
//length should be max size + 1
char mqtt_server[40] = "";
char mqtt_port[6] = "8080";
//char blynk_token[33] = "YOUR_BLYNK_TOKEN";
//default custom static IP
char static_ip[16] = "10.0.1.56";
char static_gw[16] = "10.0.1.1";
char static_sn[16] = "255.255.255.0";

char clockName[MAX_CLOCK_NAME_LEN] = "fastclk";
uint8_t clockChannel = DEFAULT_CLOCK_CHANNEL;

//flag for saving data
bool shouldSaveConfig = false;

// FastClock
bool isMaster = true;

Display display;
Radio radio(&display, true /*isMaster*/);
Fastclock fastclock(&display, &radio);

//callback notifying us of the need to save config
void saveConfigCallback () {
  Serial.println("Should save config");
  shouldSaveConfig = true;
}

void setupWifiConnection() {
  // The extra parameters to be configured (can be either global or just in the setup)
  // After connecting, parameter.getValue() will get you the configured value
  // id/name placeholder/prompt default length
  //**WiFiManagerParameter custom_mqtt_server("server", "mqtt server", mqtt_server, 40);
  //**WiFiManagerParameter custom_mqtt_port("port", "mqtt port", mqtt_port, 5);
  //WiFiManagerParameter custom_blynk_token("blynk", "blynk token", blynk_token, 34);

  //WiFiManager
  //Local intialization. Once its business is done, there is no need to keep it around
  WiFiManager wifiManager;

  //set config save notify callback
  wifiManager.setSaveConfigCallback(saveConfigCallback);

  //set static ip
  IPAddress _ip,_gw,_sn;
  _ip.fromString(static_ip);
  _gw.fromString(static_gw);
  _sn.fromString(static_sn);

  //wifiManager.setSTAStaticIPConfig(_ip, _gw, _sn);

  //add all your parameters here
  //**wifiManager.addParameter(&custom_mqtt_server);
  //**wifiManager.addParameter(&custom_mqtt_port);
  //wifiManager.addParameter(&custom_blynk_token);

  //reset settings - for testing
  //wifiManager.resetSettings();

  //set minimu quality of signal so it ignores AP's under that quality
  //defaults to 8%
  wifiManager.setMinimumSignalQuality();

  //sets timeout until configuration portal gets turned off
  //useful to make it all retry or go to sleep
  //in seconds
  //wifiManager.setTimeout(120);

  //fetches ssid and pass and tries to connect
  //if it does not connect it starts an access point with the specified name
  //and goes into a blocking loop awaiting configuration
  Serial.println("Starting autoConnect ...");
  if (!wifiManager.autoConnect("FastclockMasterAP", "password")) {
    Serial.println("failed to connect and hit timeout");
    delay(3000);
    //reset and try again, or maybe put it to deep sleep
    ESP.reset();
    delay(5000);
  }

  //if you get here you have connected to the WiFi
  Serial.println("connected...yeey :)");
  display.addLogMessage("WLAN connected");

  //read updated parameters
  //**strcpy(mqtt_server, custom_mqtt_server.getValue());
  //**strcpy(mqtt_port, custom_mqtt_port.getValue());
  //strcpy(blynk_token, custom_blynk_token.getValue());

  //save the custom parameters to FS
  if (shouldSaveConfig) {
    Serial.println("saving config");
    DynamicJsonBuffer jsonBuffer;
    JsonObject& json = jsonBuffer.createObject();
    //**json["mqtt_server"] = mqtt_server;
    //**json["mqtt_port"] = mqtt_port;
    //json["blynk_token"] = blynk_token;

    json["ip"] = WiFi.localIP().toString();
    json["gateway"] = WiFi.gatewayIP().toString();
    json["subnet"] = WiFi.subnetMask().toString();

    File configFile = SPIFFS.open("/config.json", "w");
    if (!configFile) {
      Serial.println("failed to open config file for writing");
    }

    json.prettyPrintTo(Serial);
    json.printTo(configFile);
    configFile.close();
    //end save
  }

  Serial.print("local ip: "); Serial.println(WiFi.localIP());
  Serial.print("gateway: "); Serial.println(WiFi.gatewayIP());
  Serial.print("subnet: "); Serial.println(WiFi.subnetMask());
}

void setup() {
  // if coming from deep sleep, we just go to sleep again

  // put your setup code here, to run once:
  Serial.begin(115200);
  Serial.println("Starting *** FastclockMasterESP8266");
  Serial.println(ESP.getResetReason());
  display.begin();
  display.showLog();

  //clean FS, for testing
  //SPIFFS.format();

  //read configuration from FS json
  Serial.println("mounting FS...");
  display.addLogMessage("mounting FS ...");

  if (SPIFFS.begin()) {
    Serial.println("mounted file system");
    display.addLogMessage("mounting FS done");
    if (SPIFFS.exists("/config.json")) {
      //file exists, reading and loading
      Serial.println("reading config file");
      display.addLogMessage("reading config file");
      File configFile = SPIFFS.open("/config.json", "r");
      if (configFile) {
        Serial.println("opened config file");
        size_t size = configFile.size();
        // Allocate a buffer to store contents of the file.
        std::unique_ptr<char[]> buf(new char[size]);

        configFile.readBytes(buf.get(), size);
        DynamicJsonBuffer jsonBuffer;
        JsonObject& json = jsonBuffer.parseObject(buf.get());
        json.printTo(Serial);
        if (json.success()) {
          Serial.println("\nparsed json");

          //**strcpy(mqtt_server, json["mqtt_server"]);
          //**strcpy(mqtt_port, json["mqtt_port"]);
          //strcpy(blynk_token, json["blynk_token"]);

          if (json["ip"]) {
            Serial.print("setting custom ip from config: ");
            //**strcpy(static_ip, json["ip"]);
            //**strcpy(static_gw, json["gateway"]);
            //**strcpy(static_sn, json["subnet"]);
            Serial.println(static_ip);
/*            Serial.println("converting ip");
            IPAddress ip = ipFromCharArray(static_ip);
            Serial.println(ip);*/
          } else {
            Serial.println("no custom ip in config");
          }
        } else {
          display.addLogMessage("failed loading json");
          Serial.println("failed to load json config");
        }
      }
    }
  } else {
    display.addLogMessage("failed mounting FS");
    Serial.println("failed to mount FS");
  }
  //end read
  display.addLogMessage(static_ip);
  display.addLogMessage(mqtt_server);
  Serial.print("static ip: "); Serial.println(static_ip);
  //Serial.println(blynk_token);
  //**Serial.print("mqtt: "); Serial.println(mqtt_server);

  // setupWifiConnection();


  radio.setClockChannel(clockChannel);
  radio.setClockName(clockName);
  radio.begin();
  fastclock.begin();
  pinMode(POWER_OFF_PIN, INPUT);
}

void checkForPowerOffRequest() {
  static unsigned long powerOffButtonPressed_ts = 0;
  static bool powerOffRequested = false;

  if (!digitalRead(POWER_OFF_PIN)) { // power off pressed
    if (!powerOffRequested) {
      Serial.println("Switching off? Keep button for 2 seconds to turn off.");
      powerOffRequested = true;
      powerOffButtonPressed_ts = millis();
    }
    if (millis() - powerOffButtonPressed_ts > 2000) {
      // pressed for longer than 2 seconds, now turn off
      ESP.deepSleep(ESP.deepSleepMax());
      delay(200);
    }
  } else {
    if (powerOffRequested) {
      powerOffRequested = false;
      Serial.println("Power off cancelled.");
      powerOffButtonPressed_ts = 0;
    }
  }
}



void loop(void)
{
  if (!display.showBootSequenceFinished(1000)) {
    return;
  }

  checkForPowerOffRequest();

  #if defined(AUTOSTART_CLOCK_AFTER_ms)
    #if AUTOSTART_CLOCK_AFTER_ms > 0
      static unsigned long autostart_ms = 0;
      static bool autostartDone = false;
      if (autostart_ms == 0) {
        autostart_ms = millis() + AUTOSTART_CLOCK_AFTER_ms;
      } else if (!autostartDone && millis() > autostart_ms) {
        autostartDone = true;
        fastclock.setClockHalted(false);
        Serial.println("Clock started automatically");
      }
    #endif
  #endif

  fastclock.loop();
  radio.loop();
  display.showDashboard();

}
