#include <Arduino.h>
#include <nRF24L01.h>
#include <RF24.h>
#include "config.h"
#include "radio.h"


static RF24 rf24(PIN_NRF24_CE, PIN_NRF24_CSN); // 10, 8
static byte addresses[][6] = {NETWORK_ADDRESS_MASTER_SEND, NETWORK_ADDRESS_MASTER_RECEIVE};

static int sendFailedCounter = 0;
static unsigned long stopTime = 0, pauseTime = 0;
static char thisClockName[MAX_CLOCK_NAME_LEN];

#define MAX_CLIENTS_MANAGED 20
static uint8_t numberOfKnownClients = 0;
static struct {
  uint8_t clientNetworkAddress;
  char clientName[MAX_CLIENT_NAME_LEN];
} client[MAX_CLIENTS_MANAGED];
static uint8_t nextClientNetworkAddress = 1;

void Radio::switchToSenderRole(void)
{
  Serial.println("*** changing to Tx role");
  rf24.openWritingPipe(addresses[1]);
  rf24.openReadingPipe(1,addresses[0]);
  rf24.stopListening();
  role = inTX;                  // Become the primary transmitter (ping out)
}

void Radio::switchToReceiverRole(void)
{
  Serial.println("*** changing to Rx role");
  rf24.openWritingPipe(addresses[0]);
  rf24.openReadingPipe(1,addresses[1]);
  rf24.startListening();
  role = inRX;                // Become the primary receiver (pong back)
}

void Radio::begin(void) {
  display->addLogMessage("Start RF24 radio");
  pinMode(PIN_NRF24_CSN, OUTPUT);
  pinMode(PIN_NRF24_CE, OUTPUT);

  rf24.begin();
  if (rf24.isChipConnected()) { display->addLogMessage("*** RF chip found"); }
  else { display->addLogMessage("*** ERROR: RF chip not found!"); }
  rf24.setChannel(clockChannel);
  rf24.setPALevel(RF24_PA_MAX);
  rf24.setDataRate(RF24_2MBPS);
  rf24.setAutoAck(0);
  //radio.setRetries(2,15); // Optionally, increase the delay between retries & # of retries
  rf24.setCRCLength(RF24_CRC_8);
  rf24.openWritingPipe(addresses[0]);
  rf24.openReadingPipe(1, addresses[1]);
  rf24.startListening();
  Serial.println("*** Check");
  //rf24.printDetails();
  rf24.powerUp();
  // @TODO: real random seed!
  //randomSeed(analogRead(0));
  //randomSeed(22);
  Serial.print("*** RF payload size="); Serial.print(rf24.getPayloadSize()); Serial.println(" bytes");
  if (rf24.testCarrier() || rf24.testRPD()) { display->addLogMessage("*** Carrier/RPD seen on radio"); }
  if (rf24.failureDetected) { display->addLogMessage("*** Radio error detected!"); }

  // Start the radio listening for data
  //rf24.startListening();
  switchToReceiverRole();
}

void Radio::broadcastMessageOnChannel(void *msg, int len, uint8_t channel) {
  Serial.print("# set channel "); Serial.println(channel);
  rf24.setChannel(channel);
  broadcastMessage(msg, len);
  Serial.print("# set channel "); Serial.println(clockChannel);
  rf24.setChannel(clockChannel); // switch back
  Serial.println("# done");
}

void Radio::broadcastMessage(void *msg, int len) {
  uint8_t *m = (uint8_t *) msg;
  Serial.print("*** Tx: ts="); Serial.print(millis());
  Serial.print(", len="); Serial.print(len);
  Serial.print(", typ=");
  Serial.print((uint8_t) m[0], HEX); Serial.print(" ("); Serial.print((char) m[0]);
  Serial.print("), to="); Serial.print((uint8_t) m[1], HEX);
  Serial.print(", from="); Serial.print((uint8_t) m[2], HEX);
  Serial.print(", msg="); Serial.print((uint8_t) m[3], HEX);
  Serial.print(", "); Serial.print((uint8_t) m[4], HEX);
  Serial.print(", "); Serial.print((uint8_t) m[5], HEX);
  Serial.print(", "); Serial.print((uint8_t) m[6], HEX);
  Serial.print(", "); Serial.print((uint8_t) m[7], HEX);
  Serial.print(", "); Serial.println((uint8_t) m[8], HEX);
  switchToSenderRole();
  Serial.print("#1");
  if (!rf24.writeFast(msg, len, true /*multicast*/)) {
      sendFailedCounter++;
      Serial.print("*** ERROR: failed to send msg type=");
      Serial.print(*((char *) msg));
      Serial.print(" for to=0x");
      Serial.println(m[1], HEX);
  }
  Serial.print("ts="); Serial.println(millis());
  //This is only required when NO ACK ( enableAutoAck(0) ) payloads are used
  if (millis() - pauseTime > 3) {
    pauseTime = millis();
    Serial.print("pauseTime="); Serial.println(pauseTime);
    //rf24.txStandBy();          // Need to drop out of TX mode every 4ms if sending a steady stream of multicast data
    //delayMicroseconds(130);     // This gives the PLL time to sync back up
  }
  stopTime = millis();
  Serial.print("stopTime="); Serial.println(stopTime);
                                      //This should be called to wait for completion and put the radio in standby mode after transmission, returns 0 if data still in FIFO (timed out), 1 if success
  //if (!rf24.txStandBy()) { sendFailedCounter += 3; } //Standby, block only until FIFO empty or auto-retry timeout. Flush TX FIFO if failed
  //radio.txStandBy(1000);              //Standby, using extended timeout period of 1 second

  Serial.print("*** send finished, failed="); Serial.println(sendFailedCounter);
  listenOnlyFor_ms(MIN_TIME_BETWEEN_TRANSMISSIONS_ms);
}

void Radio::broadcastClockAdvertisement(const char *clockName, uint8_t channel) {
  struct offerPairing_s msg;

  memset(&msg, 0, sizeof(msg));
  msg.msgType = msgType_OfferPairing;
  msg.from = MESSAGE_FROM_Master;
  msg.to = MESSAGE_TO_ALL;
  strncpy(msg.clockName, clockName, MAX_CLOCK_NAME_LEN);
  strncpy(thisClockName, clockName, MAX_CLOCK_NAME_LEN);
  msg.channel = channel;
  int msgLength = sizeof(msg);

  broadcastMessageOnChannel(&msg, msgLength, ADVERTISEMENT_CHANNEL);
}

// send pairing offering on advertisement channel
void Radio::broadcastOfferPairing(const char *clockName, uint8_t channel) {
  struct offerPairing_s msg;

  memset(&msg, 0, sizeof(msg));
  msg.msgType = msgType_OfferPairing;
  msg.from = MESSAGE_FROM_Master;
  msg.to = MESSAGE_TO_ALL;
  strncpy(msg.clockName, clockName, MAX_CLOCK_NAME_LEN);
  strncpy(thisClockName, clockName, MAX_CLOCK_NAME_LEN);
  msg.channel = channel;
  int msgLength = sizeof(msg);

  broadcastMessage(&msg, msgLength);
}

void Radio::broadcastClock(int day, int hour, int minute, int second, int msPerModelSecond) {
  struct clockMsg_s clockMsg;

  clockMsg.msgType = msgType_Clock;
  clockMsg.to = MESSAGE_TO_ALL;
  clockMsg.from = MESSAGE_FROM_Master;
  strncpy(clockMsg.clockName, thisClockName, MAX_CLOCK_NAME_LEN);
  clockMsg.day = day;
  clockMsg.hour = hour;
  clockMsg.minute = minute;
  clockMsg.second = second;
  clockMsg.speed = (msPerModelSecond >> 2) & 0xff; // divide by 4

  broadcastMessage(&clockMsg, sizeof(clockMsg));
}

void Radio::broadcastTextMessage(const char *text) {
  struct textMessage_s msg;

  memset(&msg, 0, sizeof(msg));
  msg.msgType = msgType_TextMessage;
  msg.from = MESSAGE_FROM_Master;
  msg.to = MESSAGE_TO_ALL;
  strncpy(msg.text, text, MAX_TEXT_MESSAGE_LEN);
  int msgLength = sizeof(msg);

  broadcastMessage(&msg, msgLength);
}


void Radio::ackPairingRequest(char *clientName, uint8_t clientNetworkAddress) {
  struct ackPairing_s msg;
  memset(&msg, 0, sizeof(msg));
  msg.msgType = msgType_ackPairingRequest;
  msg.from = MESSAGE_FROM_Master;
  msg.to = MESSAGE_TO_ALL;
  strncpy(&msg.clockName[0], thisClockName, MAX_CLOCK_NAME_LEN);
  strncpy(&msg.clientName[0], clientName, MAX_CLIENT_NAME_LEN);
  msg.clientNetworkAddress = clientNetworkAddress;
  int msgLength = sizeof(msg);

  broadcastMessage(&msg, msgLength);
}


// handle received messages
void Radio::handleRequestPairing(struct requestPairing_s *request) {
  int i=0;
  bool found;

  if (strncmp(&request->clientName[0], thisClockName, MAX_CLOCK_NAME_LEN) == 0) {
    // add this client to client list
    Serial.print("*** Rcv'd pairing request for "); Serial.println(&request->clientName[0]);
    found = false;
    for (i=0; i<numberOfKnownClients && !found; ++i) {
      if (strncmp(&client[i].clientName[0], &request->clientName[0], MAX_CLIENT_NAME_LEN) == 0) found = true;
    }
    if (!found) {
      // create new entry and set "i" to this new entry
      if (numberOfKnownClients < MAX_CLIENTS_MANAGED) {
        i = numberOfKnownClients;
        strncpy(&client[i].clientName[0], &request->clientName[0], MAX_CLIENT_NAME_LEN);
        client[i].clientNetworkAddress = nextClientNetworkAddress++;
        numberOfKnownClients++;
        Serial.print("Added new client to list of clients. Now know ");
        Serial.print(numberOfKnownClients);
        Serial.print(", new is ");
        Serial.println(&client[i].clientName[0]);
      } else {
        Serial.print("ERROR: Too many client pairing requests, max reached: ");
        Serial.println(MAX_CLIENTS_MANAGED);
        return;
      }
    }
    ackPairingRequest(&request->clientName[0], client[i].clientNetworkAddress);
  }
  // else ignore this message, as it is for another clock
  else {
    Serial.print("*** received pairing request, but it is for another clock named: ");
    Serial.println(&request->clientName[0]);
  }
}

void Radio::avoidChannelSwitchesFor_ms(unsigned int duration_ms) {
  long current_ts = millis();

  if (nextClockAdvertisement_ts <= current_ts + duration_ms) {
    nextClockAdvertisement_ts = current_ts + duration_ms + 1;
  }
}

void Radio::listenOnlyFor_ms(unsigned int duration_ms) {
  unsigned long current_ts = millis();

  if (nextSendMessage_ts <= current_ts + duration_ms) {
    nextSendMessage_ts = current_ts + duration_ms + 1;
  }
}

bool Radio::allowedToSendNow() {
  if (millis() >= nextSendMessage_ts) return true;
  return false;
}

bool Radio::checkClockAdvertisement() {
  unsigned long current_ts = millis();

  if (current_ts >= nextClockAdvertisement_ts) {
    broadcastClockAdvertisement(clockName, clockChannel);
    nextClockAdvertisement_ts = current_ts + timeBetweenClockAdvertisements_ms;
    return true;
  }

  return false;
}

bool Radio::checkPairingOffering() {
  unsigned long current_ts = millis();

  if (current_ts >= nextPairingOfferingAnnouncement_ts) {
    broadcastOfferPairing(clockName, clockChannel);
    nextPairingOfferingAnnouncement_ts = current_ts + timeBetweenPairingOfferingAnnouncements_ms;
    avoidChannelSwitchesFor_ms(timeForClientsToRespond_ms);
    listenOnlyFor_ms(timeForClientsToRespond_ms);
    return true;
  }
  return false;
}

void Radio::loop(void) {
  // put your main code here, to run repeatedly:
  /****************** Ping Out Role ***************************/
  if (isMaster)  {
    if (rf24.available()) {
      char buffer[32];
      rf24.read(buffer, 32);

      // Spew it
      if (buffer[0] > 0) {
        Serial.print(F("Rx, t="));
        Serial.print(buffer[0], HEX);
        Serial.print(F(", msg="));
        Serial.print(buffer[1], HEX);
        Serial.print(", ");
        Serial.print(buffer[2], HEX);
        Serial.print(", ");
        Serial.print(buffer[3], HEX);
        Serial.print(", ");
        Serial.println(buffer[4], HEX);
      }
    }

    // Tx handling
    if (allowedToSendNow()) {
      if (!checkPairingOffering()) {
        if (!checkClockAdvertisement()) {
          // nothing sent :-)
        }
      }
    }
  }



    #if 0

  /****************** Pong Back Role ***************************/

    if (!isMaster)
    {
      unsigned long got_time;

      if (radio.available()) {                                                                      // Variable for the received timestamp
        while (radio.available()) {                                   // While there is data ready
          radio.read( &got_time, sizeof(unsigned long) );             // Get the payload
        }

        radio.stopListening();                                        // First, stop listening so we can talk
        radio.write( &got_time, sizeof(unsigned long) );              // Send the final one back.
        radio.startListening();                                       // Now, resume listening so we catch the next packets.
        Serial.print(F("Sent response "));
        Serial.println(got_time);
     }
   }
   #endif
}
